from .DOMProperties import DOMProperties
from .DOMElement import DOMElement, DOMInfo
from .DOMObject import DOMObject
from .DOMDict import DOMDict
from .DOMList import DOMList
from .functions import document, parent, key, schema
